/*
 * patest1.c
 * Ring modulate the audio input with a 441hz sine wave for several seconds.
 *
 * Author: Ross Bencina <rossb@audiomulch.com>
 *
 * This program uses the PortAudio Portable Audio Library.
 * For more information see: http://www.audiomulch.com/portaudio/
 * Copyright (c) 1999-2000 Ross Bencina and Phil Burk
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * Any person wishing to distribute modifications to the Software is
 * requested to send the modifications to the original developer so that
 * they can be incorporated into the canonical version.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <stdio.h>
#include <math.h>
#include "portaudio.h"

#ifndef M_PI
#define M_PI  (3.14159265)
#endif

typedef struct {
	float sine[100];
	int phase;
	unsigned int sampsToGo;
}patest1data;


static int patest1Callback( void *inputBuffer, void *outputBuffer,
									unsigned long bufferFrames,
				                 	PaTimestamp outTime, void *userData )
{
	patest1data *data = (patest1data*)userData;
	float *in = (float*)inputBuffer;
	float *out = (float*)outputBuffer;

	unsigned long i;
	int finished = 0;

	(void) outTime;
	
	if( data->sampsToGo < bufferFrames ){
		/* final buffer... */
		
		for( i=0; i<data->sampsToGo; i++ ){
			*out++ = /* in[0] * */ data->sine[data->phase];		/* left */
			*out++ = in[1] * data->sine[data->phase++];		/* right */
			in += 2;
			if( data->phase >= 100 )
				data->phase = 0;
		}

		/* zero remainder of final buffer */
		for( ; i<bufferFrames; i++ ){
			*out++ = 0; /* left */
			*out++ = 0; /* right */
		}
		
		finished = 1;  /* Tell PA to stop calling back. */
		data->sampsToGo = 0;
	}
	else
	{
		for( i=0; i<bufferFrames; i++ ){
			*out++ = /* in[0] * */ data->sine[data->phase];		/* left */
			*out++ = in[1] * data->sine[data->phase++];		/* right */
			in += 2;
			if( data->phase >= 100 )
				data->phase = 0;
		}

		data->sampsToGo -= bufferFrames;
	}

	return finished;
}


extern int gRecordCounter;
extern int gPlayCounter;
extern int gDebug1;

int main(void);
int main(void)
{
	PortAudioStream *stream;
	PaError err;
	patest1data data;
	int i;

	Pa_Initialize();

	/* initialise sinusoidal wavetable */
	for( i=0; i<100; i++ )
	{
		data.sine[i] = (float) sin( ((double)i/100.) * M_PI * 2. );
	}
	data.phase = 0;
	data.sampsToGo = 5*44100;

	err = Pa_OpenStream(
				&stream,
				0,   			/* default input device */
				2,              /* stereo input */
				paFloat32,		/* 32 bit floating point input */
				NULL,
				Pa_GetDefaultOutputDeviceID(), /* default output device */
				2,              /* stereo output */
				paFloat32,      /* 32 bit floating point output */
				NULL,
				44100.,
				512,            
				4,
				paClipOff,      /* we won't output out of range samples so don't bother clipping them */
				patest1Callback,
				&data );

	if( err == paNoError ){
		if( stream ){
			err = Pa_StartStream( stream );
			if( err == paNoError ) 
			{
				while( Pa_StreamActive( stream ) )
				{
   					printf("gRecordCounter = %d, gPlayCounter = %d, active = %d, sampsToGo = %d\n",
						gRecordCounter, gPlayCounter, Pa_StreamActive( stream ), data.sampsToGo );					printf("PA running. Active = %d\n", Pa_StreamActive( stream ) ); fflush(stdout);
		  			Pa_Sleep(100); /* sleep until playback has finished */
		  		}
			}
			printf("Finished. Close stream.\n"); fflush(stdout);
			Pa_CloseStream( stream );
		}
	}else{

		fprintf( stderr, "An error occured while opening the portaudio stream\n" ); 
		fprintf( stderr, "Error number: %d\n", err );
	}
	Pa_Terminate();

}
